/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.util;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

/**
 * <p>
 * Simple JSON generator. This generator is optimized for simplicity, not
 * performance or richness in features, so code wanting to generate large or
 * complex JSON documents should use a different JSON generator.
 * </p>
 * 
 * <p>
 * This generator has specifically been written in order to minimize the
 * dependencies needed for generating a JSON document. It only uses the Java 8
 * SE API and the Apache Commons Lang 3 library.
 * </p>
 * 
 * <p>
 * This generator generates JSON document that comply with the JSON (ECMA-404)
 * standard. Compared to many other generator, this parser is very strict about
 * compliance. For example, number values that cannot be correctly represented
 * in JSON (e.g. {@link Double#NaN}) are enclosed in quotes.
 * </p>
 * 
 * <p>
 * This parser serializes Java objects to JSON using the following rules:
 * </p>
 * 
 * <ul>
 * <li>A Java value of <code>null</code> is serialized to the JSON value
 * <code>null</code>.</li>
 * <li>A {@link Boolean} is serialized to the JSON values <code>true</code> or
 * <code>false</code>.</li>
 * <li>A {@link Number} is serialized using its {@link Number#toString()
 * toString()} method. If the string representation returned by this method is
 * not a valid JSON number, the return value is serialized to a JSON
 * string.</li>
 * <li>A {@link String} is serialized to a JSON string.</li>
 * <li>A {@link Collection}, an array of objects, and any array of primitives is
 * serialized to a JSON array. Each array element is again subject to the
 * serialization rules stated in this list.</li>
 * <li>A {@link Map} is serialized to a JSON object. If an entries key is a
 * {@link String}, this string is serialized as a JSON string. If it is
 * <code>null</code>, it is serialized as the JSON string <code>"null"</code>.
 * If it is any other object, this object's <code>toString()</code> method is
 * called and the return value is serialized as a JSON string. Please note that
 * this generator does not take any precautions against duplicate keys in
 * objects. Such objects are allowed by the JSON specification, but they are
 * going to be refused by many parsers (including the
 * {@link SimpleJsonParser}).</li>
 * </ul>
 * 
 * @see SimpleJsonParser
 * 
 * @author Sebastian Marsching
 */
public class SimpleJsonGenerator {

    private static Pattern VALID_NUMBER_PATTERN = Pattern.compile(
            "-?(?:0|(?:[1-9][0-9]*))(?:\\.[0-9]+)?(?:[Ee][+\\-]?[0-9]+)?");

    /**
     * Returns the JSON representation of the specified object. Please refer to
     * the {@linkplain SimpleJsonGenerator class description} for details.
     * 
     * @param object
     *            object to be serialized to JSON. May be <code>null</code>.
     * @return JSON representation of the object.
     */
    public static String generate(Object object) {
        StringBuilder sb = new StringBuilder();
        generate(object, sb);
        return sb.toString();
    }

    /**
     * Generates the JSON representation of the specified object, appending it
     * to the specified string builder. Please refer to the
     * {@linkplain SimpleJsonGenerator class description} for details.
     * 
     * @param object
     *            object to be serialized to JSON. May be <code>null</code>.
     * @param appendTo
     *            string builder to which the string representation shall be
     *            appended.
     */
    public static void generate(Object object, StringBuilder appendTo) {
        Writer writer = new Writer() {
            private boolean closed = false;

            @Override
            public void write(char[] cbuf, int off, int len)
                    throws IOException {
                if (closed) {
                    throw new IOException("The writer has been closed.");
                }
                appendTo.append(cbuf, off, len);
            }

            @Override
            public void flush() throws IOException {
                if (closed) {
                    throw new IOException("The writer has been closed.");
                }
                // We do not have to to anything here.
            }

            @Override
            public void close() throws IOException {
                closed = true;
            }
        };
        int lengthBefore = appendTo.length();
        int lengthAfter = lengthBefore;
        try {
            generate(object, writer);
            lengthAfter = appendTo.length();
        } catch (IOException e) {
            // We know that our writer does not throw an IOException, so we do
            // not have to expect something like this here.
            throw new RuntimeException(e);
        } finally {
            // Reset the StringBuilder if there has been an exception.
            if (lengthAfter < appendTo.length()) {
                appendTo.delete(lengthBefore, lengthAfter);
            }
        }
    }

    /**
     * <p>
     * Generates the JSON representation of the specified object, writing it to
     * the specified reader. Please refer to the {@linkplain SimpleJsonGenerator
     * class description} for details.
     * </p>
     * 
     * <p>
     * If this method throws an exception, the serialized form may have
     * partially been written to the writer.
     * </p>
     * 
     * @param object
     *            object to be serialized to JSON. May be <code>null</code>.
     * @param writer
     *            writer to which the JSON representation of the object is
     *            written.
     * @throws IOException
     *             if the {@code writer} throws such an exception.
     */
    public static void generate(Object object, Writer writer)
            throws IOException {
        new SimpleJsonGenerator(writer).appendValue(object);
    }

    private Writer writer;

    private SimpleJsonGenerator(Writer writer) {
        this.writer = writer;
    }

    private void appendValue(Object object) throws IOException {
        if (object == null) {
            writer.append("null");
        } else if (object instanceof String) {
            appendString((String) object);
        } else if (object instanceof Number) {
            appendNumber((Number) object);
        } else if (object instanceof Map) {
            appendObject((Map<?, ?>) object);
        } else if (object instanceof Collection) {
            appendArray((Collection<?>) object);
        } else if (object instanceof Object[]) {
            appendArray(Arrays.asList((Object[]) object));
        } else if (object instanceof boolean[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((boolean[]) object)));
        } else if (object instanceof char[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((char[]) object)));
        } else if (object instanceof byte[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((byte[]) object)));
        } else if (object instanceof short[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((short[]) object)));
        } else if (object instanceof int[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((int[]) object)));
        } else if (object instanceof long[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((long[]) object)));
        } else if (object instanceof float[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((float[]) object)));
        } else if (object instanceof double[]) {
            appendArray(Arrays.asList(ArrayUtils.toObject((double[]) object)));
        } else if (object instanceof Boolean) {
            if (Boolean.TRUE.equals(object)) {
                writer.append("true");
            } else {
                writer.append("false");
            }
        } else {
            appendString(object.toString());
        }

    }

    private void appendArray(Collection<?> collection) throws IOException {
        writer.append('[');
        boolean first = true;
        for (Object member : collection) {
            if (!first) {
                writer.append(',');
            }
            appendValue(member);
            first = false;
        }
        writer.append(']');

    }

    private void appendNumber(Number number) throws IOException {
        String str = number.toString();
        // If the string representation of the number is not a valid JSON
        // number, we have to put it into a JSON string. For example, this can
        // happen for Double.NaN or when there are custom classes derived from
        // Number.
        if (VALID_NUMBER_PATTERN.matcher(str).matches()) {
            writer.append(str);
        } else {
            appendString(str);
        }
    }

    private void appendObject(Map<?, ?> map) throws IOException {
        writer.append('{');
        boolean first = true;
        for (Map.Entry<?, ?> entry : map.entrySet()) {
            if (!first) {
                writer.append(',');
            }
            appendObjectMember(entry);
            first = false;
        }
        writer.append('}');
    }

    private void appendObjectMember(Entry<?, ?> entry) throws IOException {
        Object key = entry.getKey();
        Object value = entry.getValue();
        if (key == null) {
            key = "null";
        }
        appendString(key.toString());
        writer.append(':');
        appendValue(value);
    }

    private void appendString(String string) throws IOException {
        writer.append('"');
        // We can deal with characters here instead of code-points because all
        // the code-points that need special handling are actually encoded in a
        // single character and other code-points are just copied as-is.
        for (int i = 0; i < string.length(); ++i) {
            char c = string.charAt(i);
            switch (c) {
            case '"':
                writer.append("\\\"");
                break;
            case '\\':
                writer.append("\\\\");
                break;
            case '\b':
                writer.append("\\b");
                break;
            case '\f':
                writer.append("\\f");
                break;
            case '\n':
                writer.append("\\n");
                break;
            case '\r':
                writer.append("\\r");
                break;
            case '\t':
                writer.append("\\t");
                break;
            default:
                if (c < 0x20) {
                    writer.append("\\u");
                    writer.append(StringUtils.leftPad(Integer.toString(c, 16),
                            4, '0'));
                } else {
                    writer.append(c);
                }
                break;
            }
        }
        writer.append('"');
    }

}
